{-# LANGUAGE Arrows, GADTs, PackageImports, KindSignatures, GeneralizedNewtypeDeriving #-}
module Network.Hermes.Agent where

import Prelude hiding(id,(.))
import Control.Category
import Control.Arrow
import Control.Concurrent
import Control.Applicative
import Data.Typeable
import "monads-tf" Control.Monad.State.Strict
import Data.Serialize(Serialize)

import Network.Hermes


-- In 
data Agent :: * -> * -> * where
  Pure        :: (b -> c) -> Agent b c
  Print       :: (Show s, Show b) => s -> Agent b ()
  AgentIO     :: (b -> IO c) -> Agent b c
  Composition :: Agent b t -> Agent t c -> Agent b c
  First       :: Agent b c -> Agent (b,d) (c,d)
  Second      :: Agent b c -> Agent (d,b) (d,c)
  ALeft       :: Agent b c -> Agent (Either b d) (Either c d)
  ARight      :: Agent b c -> Agent (Either d b) (Either d c)
  -- Apply       :: Agent (Agent b c, b) c
  -- * 
  Jump        :: (Serialize b, Typeable b) => Maybe Int -> Agent b () -> Agent (UUID,b) ()

jump :: (Serialize a, Typeable a) => Agent a () -> Agent (UUID,a) ()
jump = Jump Nothing

-- | There isn't much to compilation, but it must be done.
data CompiledAgent tag a = CA tag (Agent a ())
                       deriving(Show)

compileAgent :: (Serialize tag, Typeable tag) 
                => tag -- ^ A unique tag used to identify this
                      -- particular agent. For safety, it should
                      -- preferably be a data type defined in the same
                      -- module as the agent and not exported. Likely
                      -- a unit type, then.
                -> Agent a ()
                -> CompiledAgent tag a
compileAgent tag agent = CA tag $ evalState (compile agent) 1
  where
    compile :: Agent a b -> State Int (Agent a b)
    compile (Jump Nothing a) = do
      count <- get
      put $ succ count
      return $ Jump (Just count) a
    compile (Composition a b) = Composition <$> compile a <*> compile b
    compile (First a)         = First       <$> compile a
    compile (Second a)        = Second      <$> compile a
    compile (ALeft a)         = ALeft       <$> compile a
    compile (ARight a)        = ARight      <$> compile a
    compile (Pure _)          = return ()
    compile (Print _)         = return ()
    compile (AgentIO _)       = return ()
    -- The others don't have structure we need to recurse into.
    compile agent = return agent

test :: Agent (UUID,Int) ()
test = proc x -> do
  Print 42 -< (snd x)
  jump (Print 13) -< x
  Print 54 -< x

-- * Running

-- | Messages, sent to and from, to tell us of the progress done
newtype Invocation param = I param
                         deriving(Typeable,Serialize)

newtype InvocationTag tag = IT (Int,tag)
                          deriving(Typeable,Serialize)

-- | Registers an agent, so other machines can execute instances that
-- pass through this one. Don't call this twice for one agent.
registerAgent :: (Serialize t, Typeable t)
                 => Context -- ^ Your Hermes context
                 -> CompiledAgent a
                 -> IO ()
registerAgent ctx (CA tag agent) = register agent
  where register (Pure _)          = return ()
        register (Print _)         = return ()
        register (AgentIO _)       = return ()
        register (Composition a b) = register a >> register b
        register (First a)         = register a
        register (Second a)        = register a
        register (ALeft a)         = register a
        register (ARight a)        = register a
        register (Jump (Just count) agent) = do
          liftIO $ forkIO $ forever $ do
            (uuid,msg) <- recv ctx (Just )
            runAgent agent msg
          return ()

-- -- | Executes an agent. This call forks.
runAgent :: CompiledAgent b -> b -> IO ThreadId
runAgent (CompiledAgent tag agent) param = runAgent' tag param

runAgent' :: Agent 
-- runAgent agent parameter = forkIO $ run agent parameter
--     where run :: Agent b c -> b -> IO c
--           run (Pure f) p = return $ f p
--           run (Print s) p = putStr (show s) >> putStrLn (show p)
--           run (AgentIO aio) p = aio p
--           run (Composition a b) p = run a p >>= run b
--           run (First a) (b,d) = do c <- run a b; return (c,d)
--           run (Second a) (d,b) = do c <- run a b; return (d,c)
--           run (ALeft a) (Left p) = Left <$> run a p
--           run (ALeft _) p@(Right _) = return p
--           run (ARight a) p@(Left _) = return p
--           run (ARight a) (Right p) = Right <$> run a p
--           run (Jump a
          



-- * Debugging

showp :: Show a => a -> String
showp a = "(" ++ show a ++ ")"

-- | For debugging, no Read intended.
instance Show (Agent a b) where
  show (Pure _)            = "arr?"
  show (Composition a b)   = show a ++ " >>> " ++ show b
  show (First a)           = "first " ++ showp a
  show (Second a)          = "second " ++ showp a
  show (ALeft a)           = "left " ++ showp a
  show (ARight a)          = "right " ++ showp a
  -- show Apply            = "Apply"
  -- *
  show (Print str)         = "Print " ++ show str
  show (Jump Nothing a)    = "Jump ? " ++ showp a 
  show (Jump (Just int) a) = "Jump " ++ show int ++ " " ++ showp a
  show (AgentIO _)         = "IO?"

-- * Instances

instance Category Agent where
  id = Pure id
  (.) = flip Composition

instance Arrow Agent where
  arr = Pure
  first = First
  second = Second

instance ArrowChoice Agent where
  left = ALeft
  right = ARight
