{-# LANGUAGE PackageImports, RecordWildCards, TypeSynonymInstances, NamedFieldPuns, ScopedTypeVariables, 
    StandaloneDeriving, ViewPatterns #-}
module Skynet.Client(clientMain) where

import Skynet.Protocol
import Skynet.Resolver
import Skynet.Config
import Skynet.Client.State
import Skynet.Hermes
import Skynet.Util
import Prelude hiding(log, catch)
import Network.Hermes.MChan
import Skynet.Freespace

import System.Random
import Control.Arrow
import Control.Concurrent
import Control.Concurrent.STM
import "mtl" Control.Monad.State
import "mtl" Control.Monad.Reader
import Control.Applicative
import Data.Typeable
import Data.Data
import qualified Data.Map as M
import Data.Map(Map)
import System.Log.Logger
import Control.Exception
import Data.Maybe
import System.Posix.User
import System.Posix.Types
import System.FilePath

import Foreign.C.Error
import qualified System.Fuse as Fuse
import System.Fuse(FileStat(..), fuseMain, FuseOperations(..), defaultFuseOps,
                   defaultExceptionHandler, EntryType(..), OpenFileFlags(..),
                   FileSystemStats(..), SyncType(..))
import Data.ByteString(ByteString)
import qualified Data.ByteString as B

-- | Grab a lease, possibly opening a new one in the process
getLease :: URI -> LeaseMode -> Skynet (Maybe Lease)
getLease uri@URI{..} mode = do
  log DEBUG "skynet.client.getLease" "Looking up " uri
  -- First off, check for a pre-existing lease
  cachedLease <- lookupLease uriPath
  -- Get a new lease, or check an existing one for compatibility
  maybe (getNewLease True) checkLease cachedLease
  where
    insert lease = do
      cs <- askCS
      log DEBUG "skynet.client.getLease" "Inserting lease: " lease
      liftIO $ atomically $ modifyTVar (leasePath cs) (M.insert uriPath lease)
    checkLease lease@Lease{..} = do
      -- Make sure the lease is in the correct mode
      -- liftIO $ print (leaseMode,mode)
      -- liftIO $ print (leaseURIType,uriType)
      let upgradeRW = case (leaseMode,mode) of
            (ReadWrite,_) -> False
            (ReadOnly,ReadOnly) -> False
            otherwise -> True
          upgradeType = case (leaseURIType,uriType) of
            (_,URIWildcard) -> False
            (a,b) | a == b -> False
            otherwise -> True
      if upgradeRW || upgradeType
        then do
        -- We need to reopen the lease in a more permissive/specific mode.
        reply <- query (masterAddress leaseAddress) $ Reopen leaseID ReadWrite uriType
        case reply of
          Nothing -> return Nothing -- Timeout. Drat.
          Just Completed -> do
            let lease' = lease { leaseMode = mode, leaseURIType = uriType }
            insert lease' -- Reinsert lease, overwriting old version
            return $ Just lease'
          Just e -> do
            log ERROR "skynet.client.getLease" "Unexpected reply" e
            return Nothing
        else return $ Just lease
    getNewLease firstTry = do
      log INFO "skynet.client.getLease" "Opening " uri
      -- Okay, no lease. Go ask for one.
      -- TODO: Avoid asking twice on simultaneous calls. It's harmless, but still.
      coherency <- defaultCoherency <$> askCS
      timeout <- maxLeaseTimeout . base <$> ask
      maybeMaster <- fmap masterAddress <$> resolvePath uriPath
      master <- maybe (throwM eNOENT) return maybeMaster
      reply <- query master $ Open uri mode coherency timeout
      log DEBUG "skynet.client.getLease" "Open reply: " reply
      case reply of
        Just (Opened lease) -> do
          insert lease
          return $ Just lease
        Just OpenFailedWrongType -> do
          -- Well, nothing to do but deal with it. The error depends on what we were wanting.
          throwM $ case uriType of
            URIDirectory -> eNOTDIR
            URIFile -> eISDIR
            URIWildcard -> eIO
        Just (OpenFailedNotMaster address) | firstTry -> do
          -- Okay, wrong server to ask, so let's try another one.
          resolverInvalidate uriPath
          when (isJust address) $ resolverInsert uriPath (fromJust address) (uriType == URIDirectory)
          getNewLease False
        Just (OpenFailedNotMaster _) -> do
          -- Wrong server to ask - twice in a row. Oh dear.
          log ERROR "skynet.client.getLease" "Unable to look up path " uri
          return Nothing
        Nothing -> return Nothing
        Just e -> do
          log ERROR "skynet.client.getLease" "Unexpected reply: " e
          return Nothing

-- All the FUSE functions operate by reading or writing to the cache.  The
-- cacheGet/cachePut functions take care of coherency, pushing writes out to the server as
-- necessary.

-- | Depending on a passed-in lease, either:
--
-- * Throw an error (if the lease is broken)
-- * Return cached data (if the lease is Proxy and we have some)
-- * Fetch new data (if the lease is Local, or we lack data), then insert it in the cache (if Proxy)
cacheGet :: Skynet (Maybe Lease) -- ^ A function to get the lease
            -> (ClientState -> TVar (FMap a)) -- ^ Projection for the required cache segment
            -> FilePath -- ^ The path to look up
            -> (Lease -> Skynet a) -- ^ A function used to fetch the data if it isn't cached
            -> Skynet a
cacheGet leaseGetter projection path fetch = do
  lease <- leaseGetter
  -- TODO: Recover from errors
  maybe (throwM eIO) getter lease
  where
    getter lease@Lease{leaseType} | leaseType == Proxy = do
      cs <- askCS
      cached <- liftIO $ atomically $ M.lookup path <$> readTVar (projection cs)
      maybe insert return cached
      where
        insert = do
          datum <- fetch lease
          cs <- askCS
          log DEBUG "skynet.client.cacheGet" "Inserting for " path
          liftIO $ atomically $ modifyTVar (projection cs) (M.insert path datum)
          return datum
    getter lease = fetch lease -- It's a (server-)local lease, so caching is pointless

-- | Depending on the passed-in lease, either:
--
-- * Throw an error (if the lease is broken)
-- * Queue passed data and cache it (if the lease is Proxy and there's buffer space)
-- * Immediately send passed data (if the lease is Local or we're out of space)
cachePut :: Skynet (Maybe Lease)  -- ^ A function to get the lease
            -> (ClientState -> TVar (FMap a)) -- ^ Projection for the required cache segment
            -> FilePath -- ^ The path to update
            -> (Lease -> Skynet ()) -- ^ A function used to push the update to the server
            -> Int -- ^ An approximation of the memory cost of keeping this buffered (in bytes)
            -> a -- ^ The value to update the cache segment with
            -> Skynet ()
cachePut leaseGetter projection path push bufInc val = do
  -- TODO: Recover from errors, to the degree that's possible at all
  -- Update our cache
  cs <- askCS
  liftIO $ atomically $ modifyTVar (projection cs) (M.insert path val)
  -- Queue the update
  bufferPut leaseGetter push bufInc

-- | Queues a server call, or runs it immediately if there's no space
bufferPut :: Skynet (Maybe Lease) -- ^ A function to get the lease
             -> (Lease -> Skynet ()) -- ^ A function used to push the update to the server
             -> Int -- ^ An approximation of the memory cost of keeping this buffered (in bytes)
             -> Skynet ()
bufferPut leaseGetter push bufInc = do
  lease <- leaseGetter
  -- Queue (or immediately send) the update
  maybe (throwM eIO) putter lease
  where
    putter lease@Lease{leaseType} | leaseType == Proxy = do
      cs <- askCS
      -- We *can* queue the update. First, check if we have room to do so.
      bufLim <- bufferLimit . base <$> ask
      env <- ask
      (ok,startWriter) <- liftIO $ atomically $ do
        bufSz <- readTVar $ sendQueueSize cs
        -- Is there space?
        if bufSz + bufInc <= bufLim
          then do
          -- Check whether we need to start a buffer-writer
          startWriter <- not <$> existsMChan (sendQueue cs) (leaseID lease)
          -- There's space. Take it.
          modifyTVar (sendQueueSize cs) (+ bufInc)
          writeMChan' (sendQueue cs) (leaseID lease) (runReaderT (push lease) env, bufInc)
          return (True,startWriter)
          else return (False,False) -- No space.
      unless ok $ putter (lease { leaseType = Local })
      when startWriter $ startBufferWriter lease
    -- Lease is Local, or there's no space: push immediately.
    putter lease = push lease

-- | A process that continually writes out the buffers for some lease
startBufferWriter :: Lease -> Skynet ()
startBufferWriter lease@Lease{..} = do
  log INFO "skynet.bufferWriter" (show lease ++ " starting") ()
  forkTrapped ("skynet.bufferWriter for " ++ show lease) runWriter
  return ()
  where runWriter = do
          cs <- askCS
          buffer <- liftIO $ atomically $ do
            b <- readMChan (sendQueue cs) leaseID
            -- Delete the buffer if it's empty
            when (isNothing b) $ deleteMChan (sendQueue cs) leaseID
            return b
          case buffer of
            Just (writer,bufsz) -> do
              log DEBUG "hermes.client.bufferWriter" (show lease ++ " writing") ()
              liftIO writer
              log DEBUG "hermes.client.bufferWriter" (show lease ++ " wrote") ()
              liftIO $ atomically $ modifyTVar (sendQueueSize cs) (subtract bufsz)
              runWriter
            Nothing -> do
              log INFO "skynet.bufferWriter" (show lease ++ " shutting down") ()

queryEIO :: HermesID -> Message -> Skynet Message
queryEIO hid msg = do
  reply <- query hid msg
  -- TODO: Arrange for recovery instead of throwing EIO
  case reply of
    Nothing -> throwM eIO
    Just (Error err) -> throwM err
    Just m -> return m

--
-- Some utilities that have no direct FUSE counterpart
--

logEIO :: Show s => String -> s -> Skynet a
logEIO source s = do
  log ERROR source "Unexpected reply: " s
  liftIO $ throwM eIO

-- | Converts a relative path to an absolute, assuming it's relative to the root directory
-- (FUSE always gives absolute paths, but it likes to give foo/bar when it means /foo/bar)
makeAbsolute :: FilePath -> FilePath
makeAbsolute path | isRelative path = "/" </> path
makeAbsolute path = path

modeToSkynet Fuse.ReadOnly = ReadOnly
modeToSkynet _ = ReadWrite

getDirectory :: FilePath -> Skynet [FilePath]
getDirectory (makeAbsolute -> path) = do
  cacheGet (getLease (URI URIDirectory path) ReadOnly) dirCache path fetch
  where
    fetch Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (ReadDir leaseID)
      case reply of
        DirData dir -> return dir
        e -> logEIO "skynet.client.getDirectory" e

-- 
-- Below this lies the FUSE interface
-- 

setFileTimes :: FilePath -> EpochTime -> EpochTime -> Skynet ()
setFileTimes (makeAbsolute -> path) actime modtime = do
  log DEBUG "skynet.client.setFileTimes" "" (path,actime,modtime)
  fstat <- getFileStat path
  cachePut (getLease (URI URIWildcard path) ReadWrite) fstatCache path push 128
    fstat { statAccessTime = actime,
            statModificationTime = modtime
          }
  where
    push Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (UTime leaseID actime modtime)
      case reply of
        Completed -> return ()
        e -> logEIO "skynet.client.setFileTimes" e

skynetRead :: FilePath -> Lease -> ByteCount -> FileOffset -> Skynet ByteString
skynetRead (makeAbsolute -> path) lease (fromIntegral -> size) (fromIntegral -> offset) = do
  log DEBUG "skynet.client.read" "" (path,lease,size,offset)
  reply <- queryEIO (masterAddress . leaseAddress $ lease) (Read (leaseID lease) offset size)
  case reply of
    FileData bs -> do
      return bs
    e -> logEIO "skynet.client.read" e

skynetWrite :: FilePath -> Lease -> ByteString -> FileOffset -> Skynet ByteCount
skynetWrite (makeAbsolute -> path) _ bs offset = do
  bufferPut (getLease (URI URIFile path) ReadWrite) push (B.length bs)
  cs <- askCS
  liftIO $ atomically $ modifyTVar (fstatCache cs) (M.delete path)
  return $ fromIntegral $ B.length bs
  where
    push Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (Write leaseID (fromIntegral offset) bs)
      case reply of
        Completed -> return ()
        e -> logEIO "skynet.client.write" e

readSymbolicLink :: FilePath -> Skynet FilePath
readSymbolicLink (makeAbsolute -> path) = do
  log DEBUG "skynet.client.readSymbolicLink" "" path
  cacheGet (getLease (URI URIFile path) ReadOnly) linkCache path fetch
  where
    fetch Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (ReadLink leaseID)
      case reply of
        LinkData link -> return link
        e -> logEIO "skynet.client.readSymbolicLink" e

createSymbolicLink :: FilePath -> FilePath -> Skynet ()
createSymbolicLink linkTo (makeAbsolute -> path) = do
  log DEBUG "skynet.client.createSymbolicLink" "" (linkTo,path)
  cachePut (getLease (URI URIDirectory (takeDirectory path)) ReadWrite) linkCache path push maxBound linkTo
  where push Lease{..} = do
          reply <- queryEIO (masterAddress leaseAddress) (Symlink leaseID path linkTo)
          case reply of
            Completed -> do
              cs <- askCS
              liftIO $ atomically $ modifyTVar (dirCache cs) (M.adjust (takeFileName path :) (takeDirectory path))
            e -> logEIO "skynet.client.createSymbolicLink" e

createFile :: FilePath -> Fuse.EntryType -> FileMode -> a -> Skynet ()
createFile path Fuse.RegularFile mode _ = do
  -- Create a regular file. We pass this task off to the parent directory of the file.
  let parent = takeDirectory path
  address <- resolvePath parent
  mLease <- getLease (URI URIDirectory parent) ReadWrite
  maybe (throwM eIO) create mLease
  where
    create Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (Create leaseID path mode)
      case reply of
        Completed -> do
          -- Yay, file created!
          cs <- askCS
          liftIO $ atomically $ modifyTVar (dirCache cs) (M.adjust (takeFileName path :) (takeDirectory path))
        e -> logEIO "skynet.client.createFile" e

open :: FilePath -> Fuse.OpenMode -> OpenFileFlags -> Skynet Lease
open (makeAbsolute -> path) (modeToSkynet -> mode) flags = do
  log DEBUG "skynet.client.open" "" (path,mode)
  -- First, let's check if the file exists or not
  address <- resolvePath path
  case address of
    Just _ -> do
      -- It exists. Just open normally.
      mLease <- getLease (URI URIFile path) mode
      maybe (throwM eIO) return mLease    
    Nothing -> do
      -- File doesn't exist. Hmm.
      log WARNING "skynet.client.open" "Nonexistent file: " path
      throwM eNOENT

getFileStat :: FilePath -> Skynet FileStat
getFileStat (makeAbsolute -> path) = do
  log DEBUG "skynet.client.getFileStat" "" path
  cacheGet (getLease (URI URIWildcard path) ReadOnly) fstatCache path fetch
  where
    fetch Lease{..} = do
      reply <- queryEIO (masterAddress leaseAddress) (FStat leaseID)
      case reply of
        StatData stat -> do
          uid <- liftIO getRealUserID
          gid <- liftIO getRealGroupID
          return $ stat{statFileOwner = uid, statFileGroup = gid}
        e -> logEIO "skynet.client.getFileStat" e
  

createDirectory :: FilePath -> FileMode -> Skynet ()
createDirectory (makeAbsolute -> path) mode = undefined
removeLink :: FilePath -> Skynet ()
removeLink (makeAbsolute -> path) = undefined
removeDirectory :: FilePath -> Skynet ()
removeDirectory (makeAbsolute -> path) = undefined
rename :: FilePath -> FilePath -> Skynet ()
rename (makeAbsolute -> old) (makeAbsolute -> new) = undefined
setFileMode :: FilePath -> FileMode -> Skynet ()
setFileMode (makeAbsolute -> path) mode = undefined
setFileSize :: FilePath -> FileOffset -> Skynet ()
setFileSize (makeAbsolute -> path) size = undefined
getFileSystemStats :: String -> Skynet FileSystemStats
getFileSystemStats _ = undefined
flush :: FilePath -> Lease -> Skynet ()
flush (makeAbsolute -> path) handle = undefined
release :: FilePath -> Lease -> Skynet ()
release (makeAbsolute -> path) handle = return ()
synchronizeFile :: FilePath -> SyncType -> Skynet ()
synchronizeFile (makeAbsolute -> path) sync = undefined
releaseDirectory :: FilePath -> Skynet ()
releaseDirectory (makeAbsolute -> path) = undefined
synchronizeDirectory :: FilePath -> SyncType -> Skynet ()
synchronizeDirectory (makeAbsolute -> path) sync = undefined
access :: FilePath -> Int -> Skynet ()
access (makeAbsolute -> path) mode = undefined

openDirectory :: FilePath -> Skynet ()
openDirectory (makeAbsolute -> path) = do
  log DEBUG "skynet.client.openDirectory" "" path
  -- We have no way of knowing whether we'll be needing it for
  -- read-only or read-write access yet, but we can reopen
  -- the latter.
  getLease (URI URIDirectory path) ReadOnly
  return ()

readDirectory :: FilePath -> Skynet [(FilePath, FileStat)]
readDirectory (makeAbsolute -> path) = do
  log DEBUG "skynet.client.readDirectory" "" path
  -- We read the directory and stat the files all in one go. Is that
  -- the right way to go about it, I wonder?
  --
  -- Well, anything else would require rewriting HFuse. Maybe later.
  files <- getDirectory path
  stats <- mapM getFileStat files
  return $ zip files stats

clientMain :: Skynet ()
clientMain = do
  -- TODO: Start lease refresher
  -- Finally, start FUSE
  env <- ask
  let wrap :: Skynet a -> IO (Either Errno a)
      wrap f = (Right <$> runReaderT f env) `catch` (return . Left)
      wrap' :: Skynet () -> IO Errno
      wrap' f = (runReaderT f env >> return eOK) `catch` return
      fh = defaultFuseOps {
        fuseGetFileStat = \a -> wrap (getFileStat a),
        fuseReadSymbolicLink = \a -> wrap (readSymbolicLink a),
        fuseCreateDevice = \a b c d -> wrap' (createFile a b c d),
        -- fuseCreateDirectory = \a b -> wrap' (createDirectory a b),
        -- fuseRemoveLink = \a -> wrap' (removeLink a),
        -- fuseRemoveDirectory = \a -> wrap' (removeDirectory a),
        fuseCreateSymbolicLink = \a b -> wrap' (createSymbolicLink a b),
        -- fuseRename = \a b -> wrap' (rename a b),
        -- fuseSetFileMode = \a b -> wrap' (setFileMode a b),
        -- fuseSetFileSize = \a b -> wrap' (setFileSize a b),
        fuseSetFileTimes = \a b c -> wrap' (setFileTimes a b c),
        fuseOpen = \a b c -> wrap (open a b c),
        fuseRead = \a b c d -> wrap (skynetRead a b c d),
        fuseWrite = \a b c d -> wrap (skynetWrite a b c d),
        -- fuseGetFileSystemStats = \a -> wrap (getFileSystemStats a),
        -- fuseFlush = \a b -> wrap' (flush a b),
        -- fuseRelease = \a b -> runReaderT (release a b) env,
        -- fuseSynchronizeFile = \a b -> wrap' (synchronizeFile a b),
        fuseOpenDirectory = \a -> wrap' (openDirectory a),
        fuseReadDirectory = \a -> wrap  (readDirectory a)
        -- fuseReleaseDirectory = \a -> wrap' (releaseDirectory a),
        -- fuseSynchronizeDirectory = \a b -> wrap' (synchronizeDirectory a b),
        -- fuseAccess = \a b -> wrap' (access a b)
        }
  liftIO $ fuseMain fh defaultExceptionHandler


-- TODO: The below should really go in Skynet.Client.State, but then
-- we run into mutually recursive modules. Look into writing .hs-boot
-- file?

askCS :: Skynet ClientState
askCS = clientState <$> ask

-- | Lookup leases
class LookupLease a where
  lookupLease :: a -> Skynet (Maybe Lease)

lookupLease_ :: (Ord k) => (ClientState -> TVar (Map k a)) -> k -> Skynet (Maybe a)
lookupLease_ projection key = do
  tvar <- projection <$> askCS
  liftIO $ atomically $ M.lookup key <$> readTVar tvar

-- TODO: I'm pretty sure this can be made a lot more elegant somehow.
instance LookupLease FilePath where
  lookupLease = lookupLease_ leasePath

instance LookupLease HandleID where
  lookupLease = lookupLease_ leaseHandle

-- | Get a new handle ID
newHandle :: Skynet HandleID
newHandle = do
  cs <- askCS
  liftIO $ atomically $ do
    next <- readTVar $ nextHandle cs
    modifyTVar (nextHandle cs) succ
    return next
