{-# LANGUAGE PackageImports, RecordWildCards #-}
module Skynet.Config(
  Config(..), SkynetState(..), Skynet,
  newConfig, readConfig,
  takeRequest, didRequest,
  takeDisk, didDisk,
  forkTrapped, finally, onException,
  module Control.Monad.Reader,
  module Control.Monad.Trans
  ) where

import Skynet.Protocol
import Skynet.Util
import Skynet.Client.State(ClientState, blankClientState)
import System.Log.Logger

import Control.Applicative hiding(Alternative(..))
import Control.Concurrent.STM
import Control.Concurrent
import Data.Word
import Data.Default
import "mtl" Control.Monad.Reader hiding(lift,liftIO,MonadIO,MonadTrans)
import "mtl" Control.Monad.Trans
import qualified Network.Hermes as Hermes
import System.Log.Logger
import qualified Control.Exception as E
import Control.Concurrent.MVar(MVar)
import Happstack.State(TxControl)
import Text.ParserCombinators.Parsec

-- | Global Skynet state - the parts that are not meant to be
-- serialized, most is in some module-local happstack-state structure.
data SkynetState = SkynetState {
  base :: Config
  ,diskProcs :: TVar Int
  ,reqProcs :: TVar Int
  ,hermesCtx :: Hermes.Context
  ,hermesAuth :: Maybe Hermes.Authority
  ,txControl :: MVar TxControl
  ,clientState :: ClientState
  }

-- | Read-only configuration
data Config = Config {
  diskLimit, requestLimit :: Int
  ,maxLeaseTimeout :: Timeout
  ,defaultQueryTimeout :: Timeout
  ,maxTransferSize :: Word64
  ,skynetLogLevel, hermesLogLevel :: Priority
  ,checkpointInterval :: Int -- ^ In seconds
  ,hermesPort :: Int
  ,reliability :: Int
  ,bufferLimit :: Int
  }

instance Default Config where
  def = Config 2 8 90000 120000 65536 WARNING ALERT 3600 8378 0 (100*1024^2)

instance Show Config where
  show Config{..} = unlines [
    "# Skynet configuration.",
    "",
    "# The number of simultaneous disk requests",
    "diskLimit = " ++ show diskLimit,
    "# The number of simultaneous network requests",
    "requestLimit = " ++ show requestLimit,
    "# The maximum time leases can last before renewal, in milliseconds",
    "maxLeaseTimeout = " ++ show maxLeaseTimeout,
    "# The time allowed before giving up a request, in milliseconds",
    "defaultQueryTimeout = " ++ show defaultQueryTimeout,
    "# The transfer block size",
    "maxTransferSize = " ++ show maxTransferSize,
    "# The Skynet log level",
    "# DEBUG      Debug messages",
    "# INFO       Information",
    "# NOTICE     Normal runtime conditions",
    "# WARNING    General Warnings",
    "# ERROR      General Errors",
    "# CRITICAL   Severe situations",
    "# ALERT      Take immediate action",
    "# EMERGENCY  System is unusable ",
    "skynetLogLevel = " ++ show skynetLogLevel,
    "# The Hermes (middleware) log level",
    "hermesLogLevel = " ++ show hermesLogLevel,
    "# The checkpoint interval, in seconds",
    "checkpointInterval = " ++ show checkpointInterval,
    "# The TCP port used",
    "hermesPort = " ++ show hermesPort,
    "# Node reliability. Higher numbers are preferred for storage,", 
    "# the system will warn the user if a node with reliability 0 is used.",
    "reliability = " ++ show reliability,
    "# Maximum client buffer size, in bytes (approximate)",
    "bufferLimit = " ++ show bufferLimit
    ]

-- | Parses a Config file
readConfig :: String -> Config
readConfig = either (error . show) id . runParser parser def "<readConfig>"
  where parser = many (comment <|> statement) >> getState
        comment = try (spaces >> char '#') >> anyChar `manyTill` newline >> return ()
        statement = disklimit <|> reqlimit <|> maxlt <|> defqt <|> maxts <|> skyll <|>
                    hermll <|> checkint <|> hermport <|> rel <|> buflim
        reader :: Read a => String -> GenParser Char Config a
        reader s = do
          try $ string s >> spaces >> string "=" >> spaces
          read <$> anyChar `manyTill` newline
        -- TODO: Make these parsers look better, with lenses. First invent lenses.
        disklimit = do
          l <- reader "diskLimit"
          updateState (\s -> s { diskLimit = l })
        reqlimit = do
          l <- reader "requestLimit"
          updateState (\s -> s { requestLimit = l })
        maxlt = do
          lt <- reader "maxLeaseTimeout"
          updateState (\s -> s { maxLeaseTimeout = lt})
        defqt = do
          qt <- reader "defaultQueryTimeout"
          updateState (\s -> s { defaultQueryTimeout = qt })
        maxts = do
          mts <- reader "maxTransferSize"
          updateState (\s -> s { maxTransferSize = mts })
        skyll = do
          ll <- reader "skynetLogLevel"
          updateState (\s -> s { skynetLogLevel = ll })
        hermll = do
          ll <- reader "hermesLogLevel"
          updateState (\s -> s { hermesLogLevel = ll })
        checkint = do
          ci <- reader "checkpointInterval"
          updateState (\s -> s { checkpointInterval = ci })
        hermport = do
          port <- reader "hermesPort"
          updateState (\s -> s { hermesPort = port })
        rel = do
          r <- reader "reliability"
          updateState (\s -> s { reliability = r })
        buflim = do
          r <- reader "bufferLimit"
          updateState (\s -> s { bufferLimit = r })

type Skynet a = ReaderT SkynetState IO a

-- | Fork a thread such that crashes are logged
forkTrapped :: String -> Skynet () -> Skynet ThreadId
forkTrapped source act = do
  env <- ask
  liftIO $ forkIO $ traplogging source CRITICAL ("thread crashed: " ++ source) (runReaderT act env)

-- | A straight lift of finally
finally :: Skynet a -> Skynet b -> Skynet a
finally a b = do
  env <- ask
  liftIO $ E.finally (runReaderT a env) (runReaderT b env)

-- | A straight lift of onException
onException :: Skynet a -> Skynet a -> Skynet a
onException a b = do
  env <- ask
  liftIO $ E.onException (runReaderT a env) (runReaderT b env)

-- | Create a SkynetState from something approaching scratch
newConfig :: Hermes.Context -> Maybe Hermes.Authority -> Config -> MVar TxControl -> IO SkynetState
newConfig hermesCtx hermesAuth base txControl = do
  diskProcs <- newTVarIO 0
  reqProcs <- newTVarIO 0
  clientState <- blankClientState
  return SkynetState{..}


-- | Block until a free request process is available
takeRequest :: Skynet ()
takeRequest = do
  cur <- reqProcs <$> ask
  max <- requestLimit . base <$> ask
  takeSem cur max

-- | Make a request process available
didRequest :: Skynet ()
didRequest = dropSem =<< reqProcs <$> ask

-- | Block until a free disk process is available
takeDisk :: Skynet ()
takeDisk = do
  cur <- diskProcs <$> ask
  max <- diskLimit . base <$> ask
  takeSem cur max

-- | Make a disk process available
didDisk :: Skynet ()
didDisk = dropSem =<< diskProcs <$> ask

takeSem :: TVar Int -> Int -> Skynet ()
takeSem cur limit = liftIO $ atomically $ do
  curV <- readTVar cur
  check (curV < limit)
  modifyTVar cur succ

dropSem :: TVar Int -> Skynet ()
dropSem cur = liftIO $ atomically $ modifyTVar cur pred
