{-# LANGUAGE TemplateHaskell, DeriveDataTypeable #-}
module Skynet.Freespace(Space(..), freePercent, runFreespaceGossiper, readSpace,getSpace) where

import Skynet.Hermes
import Skynet.Config
import Skynet.Util
import Prelude hiding(log)

import Data.Derive.Serialize
import Data.DeriveTH
import Data.Serialize
import Data.Typeable
import Data.Word(Word64)
import Control.Applicative
import Control.Concurrent
import Data.List(sortBy)
import Data.Function

import System.Process

-- | Free/total space, in bytes
data Space = Space {
  spaceFree :: Word64,
  spaceTotal :: Word64,
  spaceReliability :: Int
  } deriving(Typeable,Show)

derives [makeSerialize] [''Space]

-- | Read free/total filesystem space via stat
getSpace :: Skynet Space
getSpace = do
  -- TODO: Call statfs(2) instead
  [blocksz,free,total] <- map read . words <$> liftIO (readProcess "stat" ["-f","-c","%S %a %b","."] "")
  rel <- reliability . base <$> ask
  return $ Space (free*blocksz) (total*blocksz) rel

freePercent :: Space -> Double
freePercent (Space free total _) = fromIntegral free / fromIntegral total

-- | The free-space periodic gossiper
runFreespaceGossiper :: Skynet ()
runFreespaceGossiper = do
  -- Loop, updating the factoid whenever it changes too much
  loop (Space 10 1 undefined) -- 1000% free space! (Right..)
  where loop last = do
          -- Write free space to network, if it's changed more than 1%
          now <- getSpace
          if freePercent now < freePercent last - 1
             || freePercent now > freePercent last + 1
            then do
            log NOTICE "skynet.freespace" "Updating free space to " now
            writeFactoid now () Nothing
            loop' now
            else loop' last
        loop' v = do
          -- Repeat every 30 seconds
          liftIO $ threadDelay 30000000
          loop v

-- | Returns the network's free space, ordered by reliability and then free-space
-- percentage. Won't return nodes with less than 5% free space.
readSpace :: Skynet [(HermesID,Space)]
readSpace = sortBy (sorter `on` snd) . filter (enoughSpace . snd) <$> readFactoids ()
  where sorter (Space _ _ r1) (Space _ _ r2) | r1 /= r2 = r1 `compare` r2
        sorter space1 space2 = freePercent space2 `compare` freePercent space1
        enoughSpace space = freePercent space >= 0.05
