{-# LANGUAGE ViewPatterns, ScopedTypeVariables #-}
-- | A Hermes wrapper that grabs its context from the Skynet reader monad
module Skynet.Hermes(
  send, send', recv, recv', query, query', myHermesID,
  writeFactoid, readFactoid, readFactoids, addCallback, setPeriod,
  HermesID, acceptType, refuseType
  ) where

import Skynet.Config
import Skynet.Protocol

-- import qualified Network.Hermes as Hermes
import qualified Network.Hermes as H
import Network.Hermes(HermesID)
import Control.Applicative
import Data.Serialize
import Data.Typeable
import System.Random
import System.Timeout

gc :: Skynet H.Context
gc = hermesCtx <$> ask

myHermesID :: Skynet HermesID
myHermesID = do
  ctx <- gc
  return $ H.uuid ctx

send :: (Serialize msg, Typeable msg)
         => HermesID -> msg -> Skynet ()
send uuid msg = do
  ctx <- gc
  liftIO $ H.send ctx uuid msg

send' :: (Serialize msg, Typeable msg, Serialize tag, Typeable tag)
         => HermesID -> msg -> tag -> Skynet ()
send' uuid msg tag = do    
  ctx <- gc
  liftIO $ H.send' ctx uuid msg tag

recv :: (Serialize msg, Typeable msg)
        => Skynet (HermesID,msg)
recv = do
  ctx <- gc
  liftIO $ H.recv ctx

recv' :: (Serialize msg, Typeable msg, Serialize tag, Typeable tag)
        => tag -> Skynet (HermesID,msg)
recv' tag = do
  ctx <- gc
  liftIO $ H.recv' ctx tag


-- | If you wish to queue messages without immediately calling recv, use this.
--
-- acceptType is idempotent.
acceptType :: forall tag msg. (Typeable msg, Serialize tag, Typeable tag)
              => msg -- ^ The message type to accept. Only the type is used, so undefined is fine.
              -> tag
              -> Skynet ()
acceptType msg tag = do
  ctx <- gc
  liftIO $ H.acceptType ctx msg tag

-- | If you wish to *stop* queueing messages of a given type, use this.
--
-- Calling refuseType will cause all recv calls to this type/tag
-- combination to throw RecvCancelled.
--
-- refuseType is idempotent.
refuseType :: forall tag msg. (Typeable msg, Serialize tag, Typeable tag)
              => msg -- ^ The message type to accept. Only the type is used, so undefined is fine.
              -> tag
              -> Skynet ()
refuseType msg tag = do
  ctx <- gc
  liftIO $ H.refuseType ctx msg tag


writeFactoid :: forall factoid tag. (Typeable factoid, Serialize factoid, Typeable tag, Serialize tag)
                => factoid -> tag
                -> Maybe H.TTL -- ^ The timeout, in seconds
                -> Skynet ()
writeFactoid factoid tag ttl = do
  ctx <- gc
  liftIO $ H.writeFactoid ctx factoid tag ttl

readFactoid :: forall factoid tag. (Typeable factoid, Serialize factoid, Typeable tag, Serialize tag)
               => tag -> HermesID -> Skynet (Maybe factoid)
readFactoid tag source = do
  ctx <- gc
  liftIO $ H.readFactoid ctx tag source

readFactoids :: forall factoid tag. (Typeable factoid, Serialize factoid, Typeable tag, Serialize tag)
                => tag -> Skynet [(HermesID,factoid)]
readFactoids tag = do
  ctx <- gc
  liftIO $ H.readFactoids ctx tag

addCallback :: forall msg tag. (Serialize tag, Typeable tag, Serialize msg, Typeable msg)
               => (HermesID -> tag -> msg -> Skynet ()) -> Skynet ()
addCallback function = do
  ctx <- gc
  env <- ask
  liftIO $ H.addCallback ctx (\hid tag msg -> runReaderT (function hid tag msg) env)

setPeriod :: Double -- ^ The period, in seconds
             -> Skynet ()
setPeriod period = do
  ctx <- gc
  liftIO $ H.setPeriod ctx period

-- | Convert timeouts to something we can pass to System.Timeout. The
-- max value is already effectively infinity; there's no real need to
-- call it more than once if we hit that, but we should avoid having
-- it wrap.
clamp :: Timeout -> Int
clamp time = if asInteger > toInteger (maxBound :: Int) then maxBound else fromInteger asInteger
  where asInteger = toInteger time * 1000

queryCore :: Maybe Timeout -> HermesID -> Message -> Skynet (Maybe Message)
queryCore reqTime hermesid msg = do
  time <- clamp <$> maybe (defaultQueryTimeout . base <$> ask) return reqTime
  ctx <- gc
  messageID <- liftIO randomIO
  liftIO $ H.acceptType ctx msg messageID
  send hermesid (MessageWrapper messageID msg)
  ret <- fmap snd <$> (liftIO $ timeout time (H.recv' ctx messageID))
  liftIO $ H.refuseType ctx msg messageID
  return ret

-- | Sends a wrapped Message, then waits for the reply or a timeout.
query' :: Timeout -> HermesID -> Message -> Skynet (Maybe Message)
query' time hid msg = queryCore (Just time) hid msg

-- | As query', uses the default timeout.
query :: HermesID -> Message -> Skynet (Maybe Message)
query hid msg = queryCore Nothing hid msg
