{-# LANGUAGE DeriveDataTypeable, StandaloneDeriving #-}
-- * The Skynet network protocol
module Skynet.Protocol.Base where

import System.IO(IOMode(..))
import Data.ByteString(ByteString)
import Data.Word(Word64)
import Foreign.C.Error(Errno(..))
import System.Fuse(FileStat(..))
import Data.Typeable
import Network.Hermes(HermesID)
import Happstack.State
import System.Posix.Types(FileMode,EpochTime)
import Control.Exception

data LeaseMode = ReadOnly | ReadWrite
               deriving(Eq,Typeable,Show,Ord)
instance Version LeaseMode

data Coherency = POSIX | Session | Eventual
               deriving(Typeable,Show,Eq)
instance Version Coherency

type LeaseID = Word64

type FileID = Integer

data LeaseType = Local | Proxy
               deriving(Eq,Typeable,Show)
instance Version LeaseType

data Lease = Lease { leaseID      :: LeaseID
                   , leaseType    :: LeaseType
                   , leaseURIType :: URIType
                   , leaseTimeout :: Timeout
                   , leaseMode    :: LeaseMode
                   , leaseAddress :: Address
                   }
           deriving(Eq,Typeable,Show)
instance Version Lease

-- | In milliseconds
type Timeout = Word64

data URIType = URIFile | URIDirectory | URIWildcard
             deriving(Eq,Typeable,Show)
instance Version URIType

data URI = URI { uriType :: URIType
               , uriPath :: FilePath }  
           deriving(Eq,Typeable,Show)
instance Version URI

-- | A serial number, plus the message itself
--
-- Reply messages are bare Messages, using the serial number for a tag
data MessageWrapper = MessageWrapper Word64 Message
                    deriving(Typeable,Show)

-- | Normal client<->server messages
data Message =
  Completed -- ^ Catch-all message for replies that need no data
  | NoSuchLease -- ^ Catch-all error for all requests for an unknown lease
  | Error Errno -- ^ Catch-all error reply
  -- Client->Server: Leases
  | Create LeaseID FilePath FileMode -- ^ Create a file; reply Completed. Requires a lease on the parent dir.
  | CreateCheck FilePath FileMode -- ^ Create a file; server-server request
  | CreateAbort FilePath FileMode -- ^ Server-server transaction abort
  | CreateComplete FilePath FileMode -- ^ Server-server transaction complete
  | Open URI LeaseMode Coherency Timeout -- ^ Returns a lease, used on files and
                                         -- directories. Reply Opened.
  | Close LeaseID -- ^ Close a lease. No reply.
  | KeepLease LeaseID Timeout -- ^ To reset the timeout. Reply KeptLease
  | Reopen LeaseID LeaseMode URIType -- ^ To switch from read-only to read-write or vice
                                     -- versa, or from URIWildcard to
                                     -- File/Directory. Reply Completed.
  -- Lease demotion/promotion (proxy<->local)
  | DemoteLease LeaseID -- ^ Sent by master when switching to local mode. Reply DemotedLease.
  | DemotedLease -- ^ Sent by client as response after fully flushing
  | PromoteLease LeaseID -- ^ Sent by master. No reply.
  -- Client->Server: POSIX calls. More or less.
  | Read { readLease :: LeaseID, readOffset, readSize :: Word64 } -- ^ Reply FileData (possibly multiple), then ReadDone
  | ReadDone
  | Write { writeLease :: LeaseID, writeOffset :: Word64, writeBytes :: ByteString } -- ^ Reply Completed
  | Truncate LeaseID Word64 -- ^ Reply Completed
  | UTime { utimeLease :: LeaseID, accessTime, modTime :: EpochTime } -- ^ Reply Completed
  | FStat LeaseID
  | Chmod LeaseID FileMode
  | ReadLink LeaseID -- ^ Reply LinkData
  | ReadDir LeaseID -- ^ Reply DirData. Returns the complete directory contents. Hopefully it's not too large.
  | MKDir LeaseID FilePath -- ^ Lease: parent dir. Parameter is the subdir name, not the full path
  | Unlink LeaseID -- ^ Works for both files and directories
  | Symlink { symLease :: LeaseID, symFrom, symTo :: FilePath } -- ^ Lease: Parent dir of 'symFrom'.
  -- Server->Client: Replies
  | Opened Lease
  | OpenFailedNotMaster (Maybe Address) -- ^ We're not the master. We might know who is.
  | OpenFailedWrongType -- ^ Tried to open directory as file or vice versa.
  | KeptLease LeaseType
  | FileData ByteString -- ^ Read
  | StatData FileStat   -- ^ FStat
  | LinkData FilePath -- ^ ReadLink
  | DirData  [FilePath] -- ^ ReadDir. Note how the only thing a directory knows is its contents.
  | LeaseTimeout Timeout -- ^ Open/KeepLease: What timeout is actually being used
  deriving(Typeable,Show)

deriving instance Show Errno
-- | Making Errno an Exception is somewhat iffy, but it should be fine.
deriving instance Typeable Errno
instance Exception Errno

-- | Address resolution protocol
--
-- Messages are untagged, except the final reply that contains the
-- data we originally wanted; that's tagged with the FilePath (ResolverTag).
data ResolverMessage =
  ResolverQuery FilePath FilePath HermesID -- ^ Look up the second path, believing the receiving node is responsible for the first path, send replies to <id>.
  | ResolverIntermediate FilePath Address Bool -- ^ The bool is true if the path is a directory
  | ResolverInvalidate FilePath (Maybe (Address, Bool)) -- ^ Invalidate, optionally with new data
  | ResolverReply Address Bool -- ^ Bool is true if the path is a directory, path's in the tag
  | ResolverNonexistent -- ^ Negative reply
  deriving(Eq,Typeable,Show)

newtype ResolverTag = ResolverTag FilePath
                    deriving(Typeable,Show)

data Address = Address { masterAddress :: HermesID, slaveAddresses :: [HermesID] }
             deriving(Eq,Typeable, Show)

instance Version Address
