-- | The resolver turns pathnames into node addresses, caching as
-- possible. See the Distribution Layer subsection for details.
--
-- TODO: Prune the cache after a while.
{-# LANGUAGE PackageImports, TemplateHaskell, MultiParamTypeClasses, FlexibleContexts, DeriveDataTypeable, 
             TypeFamilies, ViewPatterns, PatternGuards, FlexibleInstances #-}
module Skynet.Resolver(
  resolvePath, resolverLoop, resolverInvalidate, resolverInsert, DirCache
  ) where

import Skynet.Config
import Skynet.Hermes hiding(query)
import Skynet.Protocol
import Skynet.Storage
import Skynet.Util
import Prelude hiding(log)

import Control.Concurrent.MVar(MVar)
import Happstack.State
import "mtl" Control.Monad.State
import "mtl" Control.Monad.Reader
import Control.Applicative
import Data.Typeable
import Data.Data
import System.Log.Logger

import Data.Map(Map)
import qualified Data.Map as M
import Control.Concurrent.STM
import System.IO.Unsafe(unsafePerformIO)
import System.FilePath
import Data.Maybe
import qualified Data.List as List
import Control.Concurrent

-- | Persistent resolver state: A pathname<->node address cache
--
-- TODO: Use something more efficient than FilePaths. This will
-- require rewiring HFuse.
--
-- TODO: Prune old cache elements
data DirCache = Directory (Maybe Address) (Map FilePath DirCache)
              | File Address
              deriving(Typeable, Show)

instance Version DirCache
deriveSerialize ''DirCache

instance Component DirCache where
  type Dependencies DirCache = End
  initialValue = Directory Nothing M.empty

lookupDeepest' :: [FilePath] -> DirCache -> (Bool,Address,FilePath)
lookupDeepest' (stripRoot -> path) cache = munge $ maybe (error "Root directory uncached") id (lookup [] path cache)
  where
    -- Exact matches
    lookup pathHere [] (Directory (Just address) _) = Just (True,address,pathHere)
    lookup pathHere [] (File address) = Just (True, address, pathHere)
    -- No exact match
    lookup _ [] _ = Nothing
    -- A file where we wanted a directory? This is not the match for us.
    lookup _ _ (File _) = Nothing
    -- After recursing, use the local value if the recursion didn't find one.
    lookup pathHere (subdir:subpath) (Directory address subdirs)
      | local <- case address of Just a -> Just (False,a,pathHere); _ -> Nothing
      = (lookup (subdir:pathHere) subpath =<< M.lookup subdir subdirs) `mplus` local
    munge (a,b,c) = (a,b,joinPath ("/":reverse c))

test_dc = e
  where
    a = Directory Nothing M.empty
    b = insertCache' a (splitDirectories "/") (Address 1 []) True
    c = insertCache' b (splitDirectories "/foo") (Address 2 []) True
    d = insertCache' c (splitDirectories "/foo/bar") (Address 3 []) True
    e = insertCache' d (splitDirectories "/foo/bar/baz") (Address 4 []) False

-- | Looks up a path in the cache, returning the deepest address we
-- know if an exact match cannot be found, along with the path of that
-- address.
--
-- Returns true if an exact match is found.
lookupDeepest :: [FilePath] -> Query DirCache (Bool,Address,FilePath)
lookupDeepest path = lookupDeepest' path <$> ask

-- | Remove a key from the cache. Any subkeys are left alone.
invalidateCache' :: DirCache -> [FilePath] -> DirCache
invalidateCache' cache (stripRoot -> path) = fst $ invalidate path cache
  where 
    -- Unconditionally remove files
    invalidate [] (File _) = (error "invalidateCache",True)
    -- Only remove empty directories
    invalidate [] (Directory _ subdirs) = (Directory Nothing subdirs, M.null subdirs)
    -- Recurse, delete subdir if empty
    invalidate (dir:subpath) (Directory address subdirs)
      | Just subdir <- M.lookup dir subdirs
      , (subdir',canDelete) <- invalidate subpath subdir
      , subdirs' <- if canDelete then M.delete dir subdirs else M.insert dir subdir' subdirs
      = (Directory address subdirs', M.null subdirs' && isNothing address) -- And delete /this/ dir if empty
    -- Directory not found, thus don't delete    
    invalidate (dir:_) old@(Directory _ subdirs)
      | Nothing <- M.lookup dir subdirs
      = (old,False)

invalidateCache :: [FilePath] -> Update DirCache ()
invalidateCache path = modify (flip invalidateCache' path)

-- | Insert a key in the cache. If you insert a file at the location
-- of an existing directory, any contents of that directory is
-- deleted; similarly, if one of the intervening directories is cached
-- as a file, it is discarded.
insertCache' :: DirCache -> [FilePath] -> Address
               -> Bool -- ^ True if the key should be a directory
               -> DirCache
insertCache' cache (stripRoot -> path) address isDir = insert path cache
  where insert [] _ | not isDir = File address
        insert [] (File _) | isDir = Directory (Just address) M.empty
        insert [] (Directory _ subdirs) = Directory (Just address) subdirs
        -- Recurse through a directory, to an existing subdir
        insert (dir:subpath) (Directory address subdirs)
          | Just subdir <- M.lookup dir subdirs
          , subdir' <- insert subpath subdir
          = Directory address (M.insert dir subdir' subdirs)
        -- Recurse through a directory, creating a new subdir
        insert (dir:subpath) (Directory address subdirs)
          | subdir <- Directory Nothing M.empty
          , subdir' <- insert subpath subdir
          = Directory address (M.insert dir subdir' subdirs)
        -- An intervening path element appears to be a file. Clear it.
        insert path (File _) = insert path (Directory Nothing M.empty)

insertCache :: [FilePath] -> Address -> Bool -> Update DirCache ()
insertCache path address isDir = modify (\cache -> insertCache' cache path address isDir)

mkMethods ''DirCache ['lookupDeepest, 'invalidateCache, 'insertCache]

                                                        
resolverInvalidate :: FilePath -> Skynet ()
resolverInvalidate (splitDirectories -> path) = update $ InvalidateCache path

resolverInsert :: FilePath -> Address -> Bool -> Skynet ()
resolverInsert (splitDirectories -> path) address isDir = update $ InsertCache path address isDir

resolvePath :: FilePath -- ^ The path to look up
               -> Skynet (Maybe Address)
resolvePath path = do
  -- Check the cache first.
  log DEBUG "skynet.resolver" "Looking up " path
  (exact,address,pathSoFar) <- liftIO $ query $ LookupDeepest (splitDirectories path)
  if exact
    then return $ Just address
    else do
    -- It's not in the cache. Look it up.
    let tag = ResolverTag path
    liftIO $ infoM "skynet.resolver" $ "Resolving " ++ show path
    myID <- myHermesID
    acceptType (undefined :: ResolverMessage) tag
    send (masterAddress address) (ResolverQuery pathSoFar path myID)
    -- Wait for the final reply. Intermediate replies are inserted by the receive loop, below.
    -- TODO: Timeouts, etc.
    reply <- recv' tag
    refuseType (undefined :: ResolverMessage) tag
    case reply of
      (_,ResolverReply address isDir) -> do
        log INFO "skynet.resolver" ("Inserting " ++ show path ++ ": ")  address
        update $ InsertCache (splitDirectories path) address isDir
        return $ Just address
      (_,ResolverNonexistent) -> do
        log NOTICE "skynet.resolver" "Nonexistent path: " path
        return Nothing

-- | This loop handles all incoming resolver requests
resolverLoop :: Skynet ()
resolverLoop = forever $ do
  (lastSource,msg) <- recv
  log DEBUG "skynet.resolver" "Incoming: " (lastSource,msg)
  forkTrapped "skynet.resolver" (handle lastSource msg)
  where
    handle _ (ResolverIntermediate path address isDir) = do
      -- Optimistically insert the address in our cache. We might have use for it later.
      update $ InsertCache (splitDirectories path) address isDir
    handle _ (ResolverInvalidate (splitDirectories -> path) newData@(~(Just (address, isDir)))) = do
      -- Remove the cache entry. SECURITY: This (and Reply) could be exploited to slow things down a lot.
      update $ InvalidateCache path
      -- And insert new data, if we have it
      when (isJust newData) $ update $ InsertCache path address isDir
    handle lastSource (ResolverQuery pathHere path source) = do
      -- First off, find the path components we have some responsibility for
      chain <- findChain path
      log DEBUG "skynet.resolver"  "Looked up chain: "  chain
      let returnIntermediate inters = forkTrapped "skynet.resolver.returnIntermediate" $ do
            forM_ inters $ \(path, _, address, isDir) -> do
              send source $ ResolverIntermediate path address isDir
      case chain of
        [] -> do -- Uh oh, this query has gone wild. Log and return
                -- invalidation, then send it to the deepest node we know.
          liftIO $ noticeM "skynet.resolver" $ "Wild request: " ++ path
          send lastSource $ ResolverInvalidate pathHere Nothing
          (_,masterAddress -> nextHop,pathSoFar) <- liftIO $ query $ LookupDeepest (splitDirectories path)
          send nextHop $ ResolverQuery pathSoFar path source
        ((path', Parent, address, isDir):inters) -> do
          -- Standard intermediate request. Pass it on, and return intermediate data.
          returnIntermediate inters
          send (masterAddress address) $ ResolverQuery path' path source
        ((path', _, address, isDir):inters) -> do
          -- Final request. Return tagged data, and of course the intermediate.
          returnIntermediate inters
          -- If we have a non-parent result and it's not the requested one, that means the
          -- resource doesn't exist.
          if path' /= path
            then send' source ResolverNonexistent (ResolverTag path)
            else send' source (ResolverReply address isDir) (ResolverTag path)
