{-# LANGUAGE RecordWildCards, ViewPatterns, PackageImports, DeriveDataTypeable, TemplateHaskell, 
     PatternGuards #-}
module Skynet.Server(serverLoop) where

import Skynet.Protocol
import Skynet.Storage
import Skynet.Hermes
import Skynet.Util
import Skynet.Config
import Skynet.Freespace
import Prelude hiding(log,catch)

import Control.Concurrent.STM
import qualified Data.Map as M
import Control.Concurrent.QSem
import Control.Monad
import Data.Word(Word64)
import Data.Maybe
import Control.Applicative
import qualified Data.ByteString as BS
import Foreign.C.Error
import Control.Exception(catch)
import System.Process
import System.Log.Logger
import System.Posix.Files
import System.FilePath

serverLoop :: Skynet ()
serverLoop = forever $ do
  takeRequest
  (source, MessageWrapper mid request) <- recv
  log DEBUG "skynet.server" ("Request " ++ show mid ++ ": ") request
  forkTrapped "skynet.server" $ (((handleRequestErrno
                                   (\msg -> do
                                       log DEBUG "skynet.server" "reply: " msg
                                       send' source msg mid)
                                   source mid request)
                                  `onException` send' source (Error eIO) mid)
                                  `finally` didRequest)

clampTimeout :: Timeout -> Skynet Timeout
clampTimeout requestedTimeout = min requestedTimeout . maxLeaseTimeout . base <$> ask

-- | Blocks until the passed leases have been either demoted to Local or have timed out.
demoteLeases :: [(LeaseID,LeaseData)] -> Skynet ()
demoteLeases conflicts = do
  outstanding <- liftIO $ newTVarIO (length conflicts)
  let queryDemote (leaseID, LD{..}) = do
        log INFO "skynet.server.Open" "Demoting lease " leaseID
        answer <- query' leaseTimeout leaseOwner $ DemoteLease leaseID
        case answer of
          Just DemotedLease -> return () -- Demoted properly
          Just r -> log WARNING "skynet.server.handleRequest.Open" "Broken reply: " r
          Nothing -> return () -- Lease timed out. TODO: Handle situations where flushing takes longer than the timeout.
        leaseDemoted leaseID
        log INFO "skynet.server.Open" "Demoted lease " leaseID
        liftIO $ atomically $ modifyTVar outstanding pred
  mapM_ (forkTrapped "skynet.server.handleRequest.Open" . queryDemote) conflicts
  -- So, right. Block until the leases are done, one way or another.
  liftIO $ atomically $ check . (<= 0) =<< readTVar outstanding

-- | HandleRequest wrapper that replies with Error if an Errno is thrown
handleRequestErrno :: (Message -> Skynet ()) -> HermesID -> Word64 -> Message -> Skynet ()
handleRequestErrno reply source mid msg = do
  env <- ask
  liftIO $ runReaderT (handleRequest reply source mid msg) env `catch` \e -> runReaderT (reply (Error e)) env

handleRequest :: (Message -> Skynet ()) -> HermesID -> Word64 -> Message -> Skynet ()

handleRequest reply source mid (Create lid path mode) = do
  -- Creating a new file..  first, figure out what its master node should be. We'll use
  -- the most reliable node, with free space as tie-breaker.
  nodes <- readSpace
  continue nodes
  where
    continue [] = do
      log ALERT "skynet.server.Create" "Out of space!" ()
      throwM eNOSPC
    continue (node:nodes) = do
      let master = fst node
      -- First, inform the new master
      r <- query master $ CreateCheck path mode
      case r of
        Just Completed -> do
          -- File created. Create directory entry too.
          ok <- createFile (Just lid) path master
          if not ok
            then do
            query master $ CreateAbort path mode
            else do
            query master $ CreateComplete path mode
          simpleReply reply (return ok)
        Just (Error e) -> throwM e
        Nothing -> do
          -- Hmm, it's not responding. Try another one.
          continue nodes
        
handleRequest reply source mid (CreateCheck path _) = do
  -- Make sure we can actually make this file
  createCheck path
  reply Completed

handleRequest reply source mid (CreateAbort _ _) = return () -- Nothing to actually do here

handleRequest reply source mid (CreateComplete path mode) = do
  myid <- myHermesID
  True <- createFile Nothing path myid
  liftIO $ system $ "touch " ++ "files" </> stripRoot' path
  liftIO $ setFileMode ("files" </> stripRoot' path) mode
  reply Completed

handleRequest reply source mid (Open uri mode coherency requestedTimeout) = do
  timeout <- clampTimeout requestedTimeout
  -- First, make sure we're the master of this file - or if not, find out who is
  (typeOK,isMaster,address) <- checkMaster uri
  case () of
    () | not isMaster -> do
      reply $ OpenFailedNotMaster address
      log WARNING "skynet.server.Open" "Failed: Not master" ()
    () | not typeOK -> do
      log WARNING "skynet.server.Open" "Failed: Wrong type" ()
      reply $ OpenFailedWrongType
    otherwise -> do
      log DEBUG "skynet.server.Open" "Opening new lease" ()
      -- The requesting node may already hold a lease on this file. If so, just return that.
      (leaseID, lease, conflicts, preExisting) <- makeLease source uri mode coherency timeout
      -- Demote any leases that conflict with this one
      demoteLeases conflicts
      -- Construct the lease and reply
      let LD{..} = lease in reply $ Opened (Lease leaseID leaseType leaseURIType 
                                              leaseTimeout mode (fromJust address))
      log INFO "skynet.server.Open"
        (if preExisting then "Reopened lease " else "Opened lease ")
        leaseID
      -- TODO Mirror the lease to this file's slaves

handleRequest reply source mid l@(Reopen leaseID mode uritype) = do
  log INFO "skynet.server" "" l
  ret <- reopenLease leaseID mode uritype
  case ret of
    Nothing -> reply NoSuchLease
    Just conflicts -> do
      -- Demote now-conflicting leases..
      demoteLeases conflicts
      -- Reconstruct the lease, and reply.
      reply Completed
      log DEBUG "Reopened lease " "" leaseID

handleRequest reply source mid (Close leaseID) = do
  -- Close a lease. Depending on the other leases, this may make it
  -- possible to promote those others.
  log INFO "skynet.server.Close" "" leaseID
  promotes <- closeLease leaseID
  mapM_ (\(hid, lid) -> forkTrapped "skynet.server.handleRequest.close" $ do
            log DEBUG "skynet.server.Close" "Promoting lease for " (hid,lid)
            send hid $ PromoteLease lid)
    promotes
  reply Completed
  -- TODO: Mirror the lease to this file's slaves

handleRequest reply source mid (KeepLease lid requestedTimeout) = do
  timeout <- clampTimeout requestedTimeout
  log INFO "skynet.server.KeepLease" "" (lid,timeout)
  mType <- keepLease lid timeout
  case mType of
    Nothing -> reply NoSuchLease
    Just typ -> reply $ KeptLease typ
  -- TODO: Mirror the lease to.. etc.

handleRequest reply source mid (Read lid offset size) = do
  -- TODO: Return data in chunks? FUSE already chunkifies reads..
  mBS <- readData lid offset size
  case mBS of
    Nothing -> reply NoSuchLease
    Just bs -> reply $ FileData bs
  -- loop offset size
  -- where loop offset size = do
  --         chunkSize <- min size . maxTransferSize . base <$> ask
  --         mData <- readData lid offset chunkSize
  --         case mData of
  --           Nothing -> reply NoSuchLease
  --           Just dat -> do
  --             reply $ FileData dat
  --             if chunkSize /= size
  --               then loop (offset + chunkSize) (size - chunkSize)
  --               else reply ReadDone

handleRequest reply source mid (Write lid offset bytes) = do
  simpleReply reply $ writeData lid offset bytes

handleRequest reply source mid (Truncate lid size) = do
  simpleReply reply $ truncateFile lid size

handleRequest reply source mid (UTime lid atime mtime) = do
  simpleReply reply $ setTimes lid atime mtime

handleRequest reply source mid (FStat lid) = do
  statdata <- stat lid
  reply $ maybe NoSuchLease StatData statdata

handleRequest reply source mid (ReadLink lid) = do
  linkdata <- readLink lid
  reply $ maybe NoSuchLease LinkData linkdata

handleRequest reply source mid (ReadDir lid) = do
  dirdata <- readDir lid
  case dirdata of
    Nothing -> reply NoSuchLease
    Just dird -> reply $ DirData dird

handleRequest reply source mid (Symlink lid from to) = do
  address <- flip Address [] <$> myHermesID
  simpleReply reply $ symlink lid from to Master address

handleRequest reply source mid req = do
  log ALERT "skynet.server" "Unimplemented function " req
  throwM eNOSYS

-- | Utility function that does the Completed/NoSuchLease reply common to so many calls
simpleReply :: (Message -> Skynet ()) -> Skynet Bool -> Skynet ()
simpleReply reply bool = do
  ok <- bool
  if ok then reply Completed else reply NoSuchLease
