{-# LANGUAGE RecordWildCards, ViewPatterns, PackageImports, DeriveDataTypeable, TemplateHaskell, 
      FlexibleInstances, MultiParamTypeClasses, TypeFamilies, FlexibleContexts, TypeOperators, 
      NamedFieldPuns, ScopedTypeVariables, TupleSections #-}
module Skynet.Storage(
  StorageState,
  LeaseData(..),
  Responsibility(..),
  makeLease,
  leaseDemoted,
  checkMaster,
  findChain,
  closeLease,
  keepLease,
  stat,
  readData,
  writeData,
  truncateFile,
  readDir,
  createRootDir,
  reopenLease,
  setTimes,
  readLink,
  symlink,
  createFile,
  createCheck
  ) where

import Skynet.Protocol(Address(..), LeaseType(..), LeaseMode(..), Coherency(..),
                       Timeout(..), LeaseID(..), URI(..), URIType(..))
import Network.Hermes(HermesID)
import Skynet.Hermes(myHermesID)
import Skynet.Util(chainMap, modifyTVar,throwM,stripRoot,stripRoot')
import Skynet.Config
import Prelude hiding(catch)

import Control.Concurrent.MVar
import Control.Concurrent.STM.TVar
import Control.Concurrent.STM
import Happstack.State hiding (Proxy)
import "mtl" Control.Monad.State
import "mtl" Control.Monad.Reader
import "MaybeT" Control.Monad.Maybe
import Control.Applicative
import Control.Arrow(first,second)
import Control.Exception(bracket_, finally,throw,IOException,catch)
import Data.Typeable
import Data.Data
import Data.List
import Data.Default
import System.Log.Logger
import Data.Maybe
import Data.ByteString(ByteString, hGet, hPut)
import Data.Word(Word64)
import System.IO
import System.IO.Unsafe(unsafePerformIO)
import System.Directory
import System.FilePath
import System.Posix(getFileStatus,fileMode,fileSize,accessTime,modificationTime,statusChangeTime,
                    EpochTime(..),FileMode(..),setFileTimes,epochTime)
import Foreign.C.Error
import qualified System.Fuse as Fuse

import Data.Map(Map)
import qualified Data.Map as M

-- | FS augmentations

data Responsibility = Master | Slave | Parent
                    deriving(Typeable,Show,Eq,Ord)
instance Version Responsibility

data Times = Times {
  timeATime, timeMTime, timeCTime :: EpochTime
  } deriving(Typeable,Show)
instance Version Times
instance Default Times where def = Times 0 0 0

data PathContents =
  File
  | Directory [FilePath] Times FileMode
  | Link FilePath Times
    deriving(Typeable,Show)
instance Version PathContents

isDir :: PathContents -> Bool
isDir (Directory _ _ _) = True
isDir _ = False

data PathData = PathData {
  responsibility :: Responsibility,
  fileNodes :: Address,
  contents :: PathContents
  }
             deriving(Typeable,Show)
instance Version PathData

newtype FSState = FSState { unFSState :: Map FilePath PathData }
                deriving(Typeable,Show)
instance Version FSState

deriveSerializeFor [''FSState, ''PathData, ''Responsibility, ''PathContents, ''Times]

instance Component FSState where
  type Dependencies FSState = End
  initialValue = FSState (M.empty)

createRootDir' :: HermesID -> Update FSState ()
createRootDir' myid = put $ FSState $ M.singleton "/" $ PathData Master (Address myid []) (Directory def def 448)

readDir' :: FilePath -> Query FSState (Maybe [FilePath])
readDir' path = join . fmap (subFiles . contents) . M.lookup path . unFSState <$> ask
  where subFiles (Directory fs _ _) = Just fs
        subFiles _ = Nothing

modifyPath :: FilePath -> (PathData -> PathData) -> Update FSState ()
modifyPath path f = modify (\(FSState s) -> FSState (M.adjust f path s))

-- | Create a new file/dir/whatever. Returns False if the spot is already taken, otherwise True
createEntry' :: FilePath -> PathData -> Update FSState Bool
createEntry' path pd = fmap isJust $ runMaybeT $ do
  let (dir,file) = splitFileName path
  parent <- MaybeT $ runQuery $ readDir' dir
  if any (== file) parent
    then fail ""
    else do
    lift $ modifyPath dir (insert file)
    modify (\(FSState s) -> FSState (M.insert path pd s))
    return (Just ())
  where insert file c@PathData{contents} = c{contents = insert' file contents}
        insert' file (Directory dir times mode) = Directory (file:dir) times mode
        insert' _ other = other

setTimes' :: FilePath -> EpochTime -> EpochTime -> Update FSState ()
setTimes' path atime mtime = modifyPath path (\pd -> pd { contents = adj (contents pd) })
  where adj File = File
        adj (Directory dir (Times _ _ ct) mode) = Directory dir (Times atime mtime ct) mode
        adj (Link path (Times _ _ ct)) = Link path (Times atime mtime ct)

lookupPath' :: FilePath -> Query FSState (Maybe PathData)
lookupPath' path = M.lookup path . unFSState <$> ask

mkMethods ''FSState ['lookupPath','readDir','createRootDir','setTimes','createEntry']

data LeaseData = LD {
  leaseType :: LeaseType,
  leaseURIType :: URIType,
  leaseDemoting :: Bool,
  leaseMode :: LeaseMode,
  leaseCoherency :: Coherency,
  leaseTimeout :: Timeout,
  leaseOwner :: HermesID,
  leasePath :: FilePath
  } deriving(Typeable,Show)
instance Version LeaseData

data Leases = Leases {
  nextID :: LeaseID,
  leases :: Map LeaseID LeaseData,
  leaseRev :: Map FilePath (Map HermesID LeaseID)
  } deriving(Typeable,Show)
instance Version Leases
deriveSerializeFor [''Leases, ''LeaseData]

instance Component Leases where
  type Dependencies Leases = End
  initialValue = Leases 0 M.empty M.empty

-- And a handle cache. It can't be serialized, thus it can't be in the
-- haddock-state structure.
{-# NOINLINE leaseHandles #-}
leaseHandles :: TVar (Map LeaseID ((Handle,MVar ()),LeaseMode))
leaseHandles = unsafePerformIO $ newTVarIO M.empty

getLease' :: LeaseID -> Query Leases (Maybe LeaseData)
getLease' lid = M.lookup lid . leases <$> ask

keepLease' :: LeaseID -> Timeout -> Update Leases (Maybe LeaseType)
keepLease' lid timeout = do
  leases' <- M.adjust keepLease' lid . leases <$> get
  modify (\Leases{..} -> Leases { leases = leases', nextID, leaseRev })
  fmap leaseType . M.lookup lid . leases <$> get
  where
    keepLease' LD{..} = LD {
      leaseType = if leaseDemoting then Local else Proxy,
      leaseDemoting = False,
      leaseURIType,
      leaseMode,
      leaseCoherency,
      leaseTimeout = timeout,
      leaseOwner,
      leasePath }

closeLease' :: LeaseID -> Update Leases [(HermesID,LeaseID)]
closeLease' lid = do
  -- First, make sure the lease actually exists
  mLD <- M.lookup lid . leases <$> get
  case mLD of
    Nothing -> do
      unsafeIOToEv $ warningM "skynet.storage.closeLease" "Spurious lease release!"
      return []
    Just LD{..} -> do
      -- Yeah, okay, it exists. Fix that.
      modify $ deleteLease leasePath leaseOwner
      -- Next, check for possible promotes
      lids <- fromJust . M.lookup leasePath . leaseRev <$> get
      possibles <- chainMap lids . leases <$> get
      -- Filter out read-write posix leases
      let rws = M.filter (\(_, LD{..}) -> leaseType == Local && leaseMode == ReadWrite && leaseCoherency == POSIX)
                possibles
      -- And read-only leases
      let ros = M.filter (\(_, LD{..}) -> leaseType == Local && leaseMode == ReadOnly && leaseCoherency == POSIX)
                possibles
      -- If we have a single read-write lease, we can promote that one.
      -- If we have no read-write leases, we can promote all the read-only ones.
      case () of () | M.size rws == 1 -> return $ (\([(hid,(lid,_))]) -> [(hid,lid)]) (M.toList rws)
                  | M.size rws > 1 -> return []
                  | otherwise -> return $ map (\(hid,(lid,_)) -> (hid,lid)) $ M.toList ros
  where deleteLease leasePath leaseOwner Leases{..} =
          Leases { nextID,
                   leases   = M.delete lid leases,
                   leaseRev = M.adjust (M.delete leaseOwner) leasePath leaseRev }

makeLease' :: HermesID -> URI -> LeaseMode -> Coherency -> Timeout
             -> Update Leases (LeaseID, LeaseData, [(LeaseID,LeaseData)], Bool)
makeLease' owner URI{..} mode coherency timeout = do
  db <- get
  -- First, make sure this owner doesn't already hold this lease
  let lids = M.findWithDefault M.empty uriPath $ leaseRev db
  case M.lookup owner lids of
    -- He did, he did already hold the lease. Just return it.
    Just lid -> return $ ((lid, , [], True) . fromJust . M.lookup lid . leases) db
    -- Okay. Construct a new lease.
    Nothing -> do
      -- Filter out the conflicters, and set them to demoting.
      let conflicts = M.filter  (\(_,LD{..}) -> leaseType == Proxy &&
                                                (leaseMode == ReadWrite || mode == ReadWrite) &&
                                                leaseCoherency == POSIX && coherency == POSIX)
                      $ chainMap lids (leases db)
          conflicts' = M.map (\(lid,ld) -> (lid,ld { leaseDemoting = True })) conflicts
          conflicts'' = M.fromList . map snd . M.toList $ conflicts'
      -- Create the new lease
      let newLease = LD { leaseType = if M.null conflicts' then Proxy else Local,
                          leaseURIType = uriType,
                          leaseDemoting = False,
                          leaseMode = mode,
                          leaseCoherency = coherency,
                          leaseTimeout = timeout,
                          leaseOwner = owner,
                          leasePath = uriPath }
      -- Update the database
      put $ Leases { nextID = succ (nextID db),
                     leases = conflicts'' `M.union` (M.insert (nextID db) newLease (leases db)),
                     leaseRev = M.insert uriPath (M.insert owner (nextID db) lids) (leaseRev db) }
      -- And return the requested data.
      return (nextID db, newLease, M.elems conflicts', False)

reopenLease' :: LeaseID -> LeaseMode -> URIType
                -> Update Leases (Maybe [(LeaseID,LeaseData)])
reopenLease' lid mode utype = runMaybeT $ do
  -- First, make sure the lease exists
  lease <- MaybeT $ M.lookup lid . leases <$> get
  let lease' = lease { leaseMode = mode, leaseURIType = utype }
  -- Find any leases that conflict with this setting and need to be demoted
  lids <- MaybeT $ fmap (M.filter (/= lid)) . M.lookup (leasePath lease') . leaseRev <$> get
  conflicts <- M.filter  (\(_,lease2) -> (leaseType lease2) == Proxy &&
                                         (leaseMode lease2 == ReadWrite || leaseMode lease == ReadWrite) &&
                                         leaseCoherency lease2 == POSIX && leaseCoherency lease == POSIX)
               . chainMap lids . leases <$> get
  -- Set them to demoting and update the database
  let conflicts'  = M.map (\(lid,ld) -> (lid,ld { leaseDemoting = True })) conflicts
      conflicts'' = M.fromList . map snd . M.toList $ conflicts'
  modify (\l -> l { leases = conflicts'' `M.union` leases l })
  -- Finally, actually update the lease and return
  let lease'' = lease' { leaseType = if M.null conflicts' then leaseType lease else Local }
  modify (\l -> l { leases = M.insert lid lease'' (leases l) })
  return (M.elems conflicts')


leaseDemoted' :: LeaseID -> Update Leases ()
leaseDemoted' lid = modify (\l -> l { leases = M.adjust demote lid (leases l) })
  where demote LD{..} = LD{ leaseType = Local,
                            leaseDemoting = False,
                            leaseMode, leaseURIType, leaseCoherency, leaseTimeout, leaseOwner, leasePath }

mkMethods ''Leases ['makeLease','leaseDemoted','closeLease','keepLease','getLease','reopenLease']

-- | Reopen a lease. Returns a list of conflicting leases that must be demoted, the same way as makeLease.
reopenLease :: LeaseID -> LeaseMode -> URIType
                -> Skynet (Maybe [(LeaseID,LeaseData)])
reopenLease lid lmod utype = update $ ReopenLease' lid lmod utype


-- | Look up a particular lease
getLease :: LeaseID -> IO (Maybe LeaseData)
getLease = query . GetLease'

-- | Grab a handle. Will return nothing if the lease doesn't exist, or the request is illegal.
getHandle :: LeaseID -> LeaseMode -> IO (Maybe (Handle,MVar ()))
getHandle lid lt = do
  mHandle <- atomically $ M.lookup lid <$> readTVar leaseHandles
  case mHandle of
    Just (h,ReadWrite) -> return $ Just h -- A writable handle permits any operation
    Just (h,ReadOnly) | lt == ReadOnly -> return $ Just h -- A read-only handle allows only reads
    Just _ -> return Nothing -- Writing to a read-only handle is illegal
    Nothing -> do
      -- This lease has no attached handle. Open one.
      ld <- query $ GetLease' lid
      case ld of
        Nothing -> return Nothing -- The lease does not in fact exist
        Just LD{leasePath,leaseMode} -> do
          infoM "skynet.storage.getHandle" leasePath
          handle <- openFile ("files" </> stripRoot' leasePath) ReadWriteMode
          mvar <- newMVar ()
          atomically $ modifyTVar leaseHandles (M.insert lid ((handle,mvar),lt))
          getHandle lid lt -- This time it'll work

withHandle :: LeaseID -> LeaseMode -> (Handle -> IO ()) -> IO Bool
withHandle lid mode act = do
  handle <- getHandle lid mode
  case handle of
    Just (h,mvar) -> withMVar mvar (\() -> act h >> return True)
    Nothing -> return False

withHandle' :: LeaseID -> LeaseMode -> (Handle -> IO a) -> IO (Maybe a)
withHandle' lid mode act = do
  handle <- getHandle lid mode
  case handle of
    Just (h,mvar) -> withMVar mvar (\() -> Just <$> act h)
    Nothing -> return Nothing

-- | Read from a file
readData :: LeaseID
            -> Word64 -- ^ Offset
            -> Word64 -- ^ Size
            -> Skynet (Maybe ByteString)
readData lid (toInteger -> offset) (fromIntegral -> size) = liftIO $ withHandle' lid ReadOnly $ \h -> do
  hSeek h AbsoluteSeek offset
  hGet h size

-- | Write to a file
writeData :: LeaseID
             -> Word64 -- ^ Offset
             -> ByteString -- ^ Data
             -> Skynet Bool
writeData lid (toInteger -> offset) bytes = liftIO $ withHandle lid ReadWrite $ \h -> do
  hSeek h AbsoluteSeek offset
  hPut h bytes

-- | Stat a file/directory
stat :: LeaseID -> Skynet (Maybe Fuse.FileStat)
stat lid = liftIO $ runMaybeT $ do
  LD{leasePath} <- MaybeT $ query $ GetLease' lid
  PathData{contents} <- MaybeT $ query $ LookupPath' leasePath
  let statEntryType = case contents of
        File -> Fuse.RegularFile
        Directory _ _ _ -> Fuse.Directory
        Link _ _ -> Fuse.SymbolicLink
  (statFileSize,
   statAccessTime,statModificationTime,statStatusChangeTime,
   statFileMode) <- case contents of
    File -> do
      status <- liftIO $ getFileStatus ("files" </> stripRoot' leasePath)
      return (fileSize status,
              accessTime status, modificationTime status, statusChangeTime status,
              fileMode status)
    Directory dir (Times at mt ct) m -> return (genericLength dir,at,mt,ct,m)
    Link l (Times at mt ct) -> return (genericLength l,at,mt,ct,511)
  let statLinkCount = fromIntegral $ case contents of
        File -> 1
        -- We're fibbing, but some programs expect link count to mean something        
        Directory (length -> len) _ _ -> len + 1
        Link _ _ -> 1
      statBlocks = ceiling $ (fromIntegral statFileSize) / 4096 -- Arbitrary but common block size
      -- these get filled in at the client
      statFileOwner = 0
      statFileGroup = 0
      statSpecialDeviceID = 0
  return $ Fuse.FileStat{..}

-- | Set the atime/mtime of a file/dir
setTimes :: LeaseID -> EpochTime -> EpochTime -> Skynet Bool
setTimes lid atime mtime = fmap isJust $ runMaybeT $ do
  LD{leasePath} <- MaybeT $ query $ GetLease' lid
  PathData{contents} <- MaybeT $ query $ LookupPath' leasePath
  case contents of
    File -> liftIO $ setFileTimes ("files" </> stripRoot' leasePath) atime mtime
    _ -> update $ SetTimes' leasePath atime mtime

-- | Read a symlink
readLink :: LeaseID -> Skynet (Maybe FilePath)
readLink lid = runMaybeT $ do
  LD{leasePath} <- MaybeT $ query $ GetLease' lid
  PathData{contents} <- MaybeT $ query $ LookupPath' leasePath
  case contents of
    Link path _ -> return path
    _ -> throwM eINVAL

-- | Create a symlink
symlink :: LeaseID -> FilePath -> FilePath -> Responsibility -> Address -> Skynet Bool
symlink lid from to resp address = fmap isJust $ runMaybeT $ do
  LD{leasePath} <- MaybeT $ query $ GetLease' lid
  now <- lift timesNow
  ok <- lift $ update $ CreateEntry' from $
        PathData resp address (Link to now)
  if not ok
    then throwM eEXIST
    else return $ Just ()

-- | Create a file
createFile :: (Maybe LeaseID) -> FilePath -> HermesID -> Skynet Bool
createFile lid path master = fmap isJust $ runMaybeT $ do
  -- We don't /need/ the lease for anything, but let's make sure it's real
  when (isJust lid) $ do
    MaybeT $ query $ GetLease' (fromJust lid)
    return ()
  -- Right. Creating.
  us <- lift myHermesID
  lift $ update $ CreateEntry' path $
    PathData (if master == us then Master else Parent) (Address master []) File
  return ()
    
-- | Create the parent directories of a file. Returns false if some intervening element is
-- a file.
createCheck :: FilePath -> Skynet ()
createCheck (joinPath . stripRoot . splitDirectories -> path) = 
  liftIO $ (createDirectoryIfMissing True ("files" </> takeDirectory path)) `catch`
  (\(e::IOException) -> do
      alertM "skynet.storage.createCheck" ("createCheck error for " ++ path ++ ": " ++ show e)
      throwM eIO)

  
-- | Return the current time (thrice)
timesNow :: Skynet Times
timesNow = liftIO $ do
  now <- epochTime
  return $ Times now now now

-- | Truncate a file
truncateFile :: LeaseID -> Word64 -> Skynet Bool
truncateFile lid (toInteger -> size) = liftIO $ withHandle lid ReadWrite $ \h -> do
  hSetFileSize h size

-- | Reads the list of files in a directory
readDir :: LeaseID -> Skynet (Maybe [FilePath])
readDir lid = liftIO $ do
  ld <- getLease lid
  maybe (return Nothing) readIt ld
  where
    readIt LD{leasePath} = query $ ReadDir' leasePath

-- | Initialize the root directory. WILL DESTROY THE FILESYSTEM!
createRootDir :: Skynet ()
createRootDir = do
  myid <- myHermesID
  update $ CreateRootDir' myid


-- | Create/lookup a lease. Returns the lease, a list of leases with
-- conflicting coherency/proxyness, and true if the requesting node
-- already had a lease on this file.
--
-- The returned, conflicting leases have leaseDemoting set to true;
-- call leaseDemoted when they have been.
makeLease :: HermesID -> URI -> LeaseMode -> Coherency -> Timeout 
            -> Skynet (LeaseID, LeaseData, [(LeaseID,LeaseData)], Bool)
makeLease owner uri mode coherency timeout = liftIO $ update $ MakeLease' owner uri mode coherency timeout

-- | Set a currently-demoting lease to demoted
leaseDemoted :: LeaseID -> Skynet ()
leaseDemoted = liftIO . update . LeaseDemoted'

-- | (typeOK,True,Just address) if we're the master, otherwise (typeOK,False, maybe address)
-- typeOK is True if uriType matches isDir, False otherwise.
checkMaster :: URI -> Skynet (Bool,Bool, Maybe Address)
checkMaster URI{..} = liftIO $ do 
  datum <- query $ LookupPath' uriPath
  return $ case datum of
    Nothing -> (True,False,Nothing) -- ^ Hmm, shades of Daily WTF. FileNotFound indeed.
    Just (PathData Master address contents) -> (match contents, True, Just address)
    Just (PathData _ address contents) -> (match contents, False, Just address)
  where
    match (Directory _ _ _) | uriType == URIFile = False
    match (Link _ _)        | uriType == URIDirectory = False
    match File              | uriType == URIDirectory = False
    match _ = True

-- | Returns the responsibility, address and isDir data of all components we
-- know something for, in order from deepest to shallowest.
findChain :: FilePath -> Skynet [(FilePath, Responsibility, Address, Bool)]
findChain path = do
  let components = map (joinPath . reverse) $ takeWhile (not . null)
                   $ iterate tail $ reverse $ splitDirectories path
  -- Check every path component, from deepest to shallowest.
  catMaybes <$> (forM components $ \component -> let
    munge PathData{..} = (component, responsibility, fileNodes, isDir contents)
    in fmap munge <$> (liftIO $ query $ LookupPath' component))

-- | Close a lease. Returns a list of other leases that can now be
-- promoted.
closeLease :: LeaseID -> Skynet [(HermesID,LeaseID)]
closeLease lid = liftIO $ do
  ret <- update $ CloseLease' lid
  -- Shut down any open handles. They'd probably be GC'd, but still.
  handle <- atomically $ do
            h <- M.lookup lid <$> readTVar leaseHandles
            modifyTVar leaseHandles (M.delete lid)
            return h
  when (isJust handle) $ do 
    takeMVar $ snd $ fst $ fromJust handle
    hClose $ fst $ fst $ fromJust handle
  -- And return
  return ret

keepLease :: LeaseID -> Timeout -> Skynet (Maybe LeaseType)
keepLease lid timeout = liftIO $ update $ KeepLease' lid timeout

-- | A sum state type, so we don't need to export (and, elsewhere,
-- handle) more than one
data StorageState = StorageState
     deriving(Typeable,Show)
instance Version StorageState
deriveSerialize ''StorageState

instance Component StorageState where
  type Dependencies StorageState = Leases :+: FSState :+: End
  initialValue = StorageState

mkMethods ''StorageState []
