{-# LANGUAGE PackageImports, ViewPatterns, TupleSections, TemplateHaskell, 
             DeriveDataTypeable, TypeOperators, TypeFamilies #-}
module Main where

import Skynet.Server
import Skynet.Client
import Skynet.Config
import Skynet.Util
import Skynet.Storage
import Skynet.Resolver
import Skynet.Freespace
import qualified Skynet.Protocol as Protocol

import Network.Hermes

import Data.Monoid
import Data.Serialize
import Control.Monad
import Control.Applicative((<$>),(*>),(<*))
import System.IO
import System.Directory
import System.Environment
import qualified Data.ByteString as B
import Data.Default
import "mtl" Control.Monad.Reader
import Data.Maybe
import System.Process
import Text.ParserCombinators.Parsec
import Control.Arrow
import Data.List.Split(splitOn)
import System.Console.Haskeline(runInputT,getInputLine)
import Happstack.State(runTxSystem,createCheckpoint,Saver(..),Version,deriveSerialize,
                       Dependencies,Component(..),(:+:),End,mkMethods,Proxy(..),shutdownSystem)
import Data.Typeable
import Control.Concurrent
import Control.Exception(evaluate)
import Control.Concurrent.STM
import System.FilePath
import System.Log.Logger

-- | The One Unified Happstack State
data HappstackState = HappstackState
                    deriving(Typeable,Show)
instance Version HappstackState
deriveSerialize ''HappstackState
instance Component HappstackState where
  type Dependencies HappstackState = StorageState :+: DirCache :+: End
  initialValue = HappstackState
mkMethods ''HappstackState []


authorityFile :: FilePath
authorityFile = "skynet.authority"

contextFile :: FilePath
contextFile = "skynet.ctx"

configFile :: FilePath
configFile = "skynet.conf"

main :: IO ()
main = do
  args <- getArgs
  case args of
    [mode] | mode == "create-network"   -> createNetwork
           | mode == "authorize-node"   -> authorizeNode
    [mode,file] | mode == "create-node" -> createNode file
    (mode:flags) | mode == "run"        -> withArgs ("-f":flags) runNode
    otherwise -> putStrLn help

-- | Create and store a Hermes authority, and the (current) root node
-- Hermes context
createNetwork :: IO ()
createNetwork = unlessState $ do
  -- Authority
  putStr "Generating authority.." >> hFlush stdout
  auth <- newAuthority
  B.writeFile authorityFile $ encode auth
  putStrLn "done"
  -- Context
  putStr "Generating hermes context.." >> hFlush stdout
  ctx <- newSignedContext auth
  B.writeFile contextFile =<< snapshotContext ctx
  putStrLn "done"
  putStrLn $ "Hermes context written to " ++ contextFile
  -- Configuration
  port <- prompt "Please enter a port to bind to: "
  writeFile configFile $ show (def { reliability = 1, hermesPort = port })
  putStrLn $ "Default configuration written to " ++ configFile
  putStrLn $ "Creating root directory"
  loadState $ do
    -- Create the root directory
    createRootDir
    -- Initialize the resolver
    resolverInsert "/" (Protocol.Address (uuid ctx) []) True

decode' :: Serialize a => B.ByteString -> a
decode' = either error id . decode

putFlush s = liftIO $ putStr s >> hFlush stdout

-- | Load Skynet's state from disk, initialize happstack, initialize Hermes.
loadState :: Skynet () -> IO ()
loadState act = do
  hasState <- and <$> mapM doesFileExist [contextFile,configFile]
  if not hasState
    then putStrLn "No hermes state detected, did you forget to run create-node?"
    else do
    putFlush "Initializing Hermes.."
    ctx <- restoreContext =<< B.readFile contextFile
    -- setDebug DEBUG
    putStrLn "done"
    putFlush "Loading Hermes authority.."
    auth <- do
      e <- doesFileExist authorityFile
      if e then Just . decode' <$> B.readFile authorityFile else return Nothing
    if isJust auth then putStrLn "done" else putStrLn "not found"
    putFlush "Loading configuration.."
    config <- readConfig <$> readFile configFile
    evaluate config
    updateGlobalLogger "skynet" $ setLevel $ skynetLogLevel config
    updateGlobalLogger "hermes" $ setLevel $ hermesLogLevel config
    putStrLn "done"
    address <- joinPort (hermesPort config) . snd <$> findAddress
    putStrLn $ "Listening to " ++ show address
    startListener ctx address Nothing
    putFlush "Initializing Happstack-state.."
    txControl <- runTxSystem (FileSaver "meta/") (Proxy :: Proxy HappstackState)
    putStrLn "done"
    state <- newConfig ctx auth config txControl
    flip runReaderT state $ do
      -- Set up a process to occasionally checkpoint the database
      interval <- (* 100000) . checkpointInterval . base <$> ask
      checkpointer <- liftIO $ forkIO $ do
        forever $ do
          threadDelay interval
          createCheckpoint txControl
      -- And, at long last, run our code.
      act `finally` do
        -- Clean up - shut down happstack-state.
        putFlush "Shutting down.."
        liftIO $ killThread checkpointer
        liftIO $ threadDelay 100000 -- FIXME: Workaround for happstack-state bug where it
                                    -- locks up if shut down too quickly.
        liftIO $ createCheckpoint txControl
        liftIO $ shutdownSystem txControl
        liftIO $ putStrLn "done"
  where
    joinPort port (IPv4 a _) = IPv4 a port
    joinPort port (IPv6 a _) = IPv6 a port

authorizeNode :: IO ()
authorizeNode = loadState $ do
  notMaster <- isNothing . hermesAuth <$> ask
  if notMaster
    then liftIO $ putStrLn "authorize-node must be run on the administrator node"
    else do
    putFlush "Creating context.."
    auth <- fromJust . hermesAuth <$> ask
    ctx <- liftIO $ newSignedContext auth
    ourCtx <- hermesCtx <$> ask
    liftIO $ atomically $ do
      Nothing <- restoreAddresses ctx =<< snapshotAddresses ourCtx
      return ()
    liftIO $ putStrLn "done"
    filename <- liftIO $ askLine "Filename: "
    liftIO $ B.writeFile filename =<< encode . (uuid ctx,) <$> snapshotContext ctx
    liftIO $ putStrLn "Now copy the authorization to the node, then run create-node"

createNode :: FilePath -> IO ()
createNode file = unlessState $ do
  putFlush "Loading authorization.."
  (root,ctxBS) <- decode' <$> B.readFile file
  ctx <- restoreContext ctxBS
  putStrLn "ok"
  B.writeFile contextFile =<< snapshotContext ctx    
  putStrLn $ "Hermes context written to " ++ contextFile
  storage <- prompt "Should this node be used for storage?"
  port <- prompt "Please enter a port to bind to: "
  writeFile configFile $ show (def { reliability = if storage then 1 else 0, hermesPort = port })
  putStrLn $ "Default configuration written to " ++ configFile
  putStrLn "Initializing resolver"
  loadState $ resolverInsert "/" (Protocol.Address root []) True

unlessState :: IO () -> IO ()
unlessState act = do
  hasState <- or <$> mapM doesFileExist [authorityFile,contextFile,configFile]
  if hasState
    then putStrLn "This directory appears to already contain Hermes state, aborting"
    else act

-- | The main event: Running a skynet node
runNode :: IO ()
runNode = loadState $ do
  -- Start the free-space gossiper
  forkTrapped "skynet.freespace" runFreespaceGossiper
  -- The server loop
  forkTrapped "skynet.server" serverLoop
  -- The resolver loop
  forkTrapped "skynet.resolver" resolverLoop
  -- And finally, the FUSE client
  clientMain

help :: String
help = unlines [
  "Usage: skynet [MODE]",
  "Take care, skynet will use the current working directory for state",
  "and file storage.",
  "",
  "Execution modes:",
  "  skynet create-network: Creates a skynet network, with this computer as ",
  "                          the network administrator",
  "  skynet authorize-node: After running create-network, creates a Hermes context",
  "                          to bootstrap other nodes. It will include the address",
  "                          of this node to facilitate bootstrapping.",
  "  skynet create-node [file]: Creates a skynet node, to connect to an existing",
  "                              network. Will need a node authorization created",
  "                              by authorize-node.",
  "  skynet run [mount]:    Runs skynet, mounting the filesystem at the given",
  "                          mount-point."
  ]



-- *
-- * * Extracting interface addresses is an unholy mess. This is my poor attempt to automate the process.
-- * * It should probably go in Hermes somewhere.
-- *

findAddress :: IO (String,Address)
findAddress = do
  addresses <- filterLoopback <$> getAddresses
  -- Try to find a single address
  -- In order of preference: Global ipv4, global ipv6, private ipv4/ipv6
  let auto = msum [tryAddress (filterIPv4 . ipv4only $ addresses),
                   tryAddress (filterIPv4 addresses),
                   tryAddress addresses]
  -- TODO: An address picker for when multiple valid addresses are detected
  maybe readAddress (\a -> showAddress a >> return a) auto
  where
    tryAddress [(dev,a)] = Just (dev,a)
    tryAddress _ = Nothing
    showAddress (dev,a) = putStrLn $ "Autodetected address: " ++ dev ++ "/" ++ tostr a
    tostr (IPv4 a _) = a
    tostr (IPv6 a _) = a
    readAddress = do
      putStrLn "Address detection failed. Please enter an interface IP address (the port is ignored)"
      putStrLn "  Examples: IPv4 \"10.0.28.3\" 0"
      putStrLn "            IPv6 \"2001:470:28::2\" 0"
      putStrLn ""
      ("",) <$> prompt "> "

-- | Parses the output of "ip addr list" for addresses
getAddresses :: IO [(String,Address)]
getAddresses = do
  output <- readProcess "ip" ["addr","list"] "" `catch` (\e -> putStrLn ("Error: " ++ show e) >> return [])
  return $ either (const []) id (parse parser "<ip addr list>" output)
  where
    flatten = concat . map (\(d,as) -> zip (repeat d) as)
    parser = flatten <$> many parseDev
    parseDev = do
      dev <- devLine
      addresses <- addressLine `manyTill` lookAhead ((devLine >> return ()) <|> eof)
      return (dev,catMaybes addresses)
    devLine = do
      digit
      skipMany digit
      char ':'
      spaces
      devName <- many alphaNum
      char ':'
      anyChar `manyTill` newline
      return devName
    addressLine = do
      spaces
      (Just <$> inet) <|> (Just <$> inet6) <|> (anyChar `manyTill` newline >> return Nothing)
    genNet s = do
      try (string s >> space)
      spaces
      v <- anyChar `manyTill` char '/'
      anyChar `manyTill` newline
      return v
    inet = flip IPv4 0 <$> genNet "inet"
    inet6 = flip IPv6 0 <$> genNet "inet6"

-- | Removes loopback and IPv6 link-local addresses
filterLoopback :: [(String,Address)] -> [(String,Address)]
filterLoopback = filter check
  where check (s,_) | s == "lo" = False
        check (_,IPv6 a _) | take 3 a == "fe8" = False -- ^ TODO: Parse properly, please!
                           | take 3 a == "fe9" = False
                           | take 3 a == "fea" = False
                           | take 3 a == "feb" = False
        check _ = True

-- | Filters out local IPv4 addresses
filterIPv4 :: [(String,Address)] -> [(String,Address)]
filterIPv4 = filter check
  where check (_,IPv4 (splitOn "." -> ((read -> a):(read -> b):_)) _) 
          | a == 10 = False
          | a == 192 && b == 168 = False
          | a == 172 && b >= 16 && b <= 31 = False
        check _ = True

ipv4only :: [(String,Address)] -> [(String,Address)]
ipv4only = filter check
  where check (_, IPv4 _ _) = True
        check _ = False
